/* Copyright (c) 2001, Todd J Martin
 * $Id: pcap_tcl.c,v 1.41 2005/06/19 01:31:45 todd Exp $
 */
#include <stdlib.h>
#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#endif
#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif
#if HAVE_SYS_TIME_H
# include <sys/time.h>
#endif
#include <errno.h>
#include <string.h>
#include <tcl.h>
#include <pcap.h>
#include "pcap_tcl.h"

/*
#ifdef __FreeBSD__
# include <sys/types.h>
# include <sys/time.h>
# include <sys/ioctl.h>
# include <net/bpf.h>
#endif
*/

#define SIZE_DATALINKARRAY 24 
static struct tclpcap_datalink {
	int dlinkValue;
	const char *shortString;
	const char *longString;
} datalinkArray[] = {
	{DLT_NULL, "DLT_NULL", "NULL"},
	{DLT_EN10MB, "DLT_EN10MB", "Ethernet"},
	{DLT_IEEE802, "DLT_IEEE802", "Token Ring"},
	{DLT_ARCNET, "DLT_ARCNET", "Arcnet"},
	{DLT_SLIP, "DLT_SLIP", "SLIP"},
	{DLT_PPP, "DLT_PPP", "PPP"},
	{DLT_FDDI, "DLT_FDDI", "FDDI"},
	{DLT_ATM_RFC1483, "DLT_ATM_RFC1483", "RFC 1483 encapsulated ATM"},
	{DLT_RAW, "DLT_RAW", "Raw IP"},
	{DLT_PPP_SERIAL, "DLT_PPP_SERIAL", "PPP HDLC"},
	{DLT_PPP_ETHER, "DLT_PPP_ETHER", "PPPoE"},
	{DLT_C_HDLC, "DLT_C_HDLC", "Cisco PPP HDLC"},
	{DLT_IEEE802_11, "DLT_IEEE802_11", "IEEE 802.11 wireless LAN"},
	{DLT_LOOP, "DLT_LOOP", "OpenBSD loopback encapsulation"},
	{DLT_LINUX_SLL, "DLT_LINUX_SLL", "Linux cooked capture encapsulation"},
	{DLT_LTALK, "DLT_LTALK", "Apple LocalTalk"},
	{DLT_LINUX_IRDA, "DLT_LINUX_IRDA", "Linux-IrDa packets"},
	{DLT_ARCNET_LINUX, "DLT_ARCNET_LINUX", "Linux ARCNET"},
	{DLT_IEEE802_11_RADIO, "DLT_IEEE802_11_RADIO", "IEEE 802.11 wireless LAN"},
	{DLT_SUNATM, "DLT_SUNATM", "SunATM device"},
	{DLT_IP_OVER_FC, "DLT_IP_OVER_FC", "RFC 2625 IP over Fibre Channel"},
	{DLT_PRISM_HEADER, "DLT_PRISM_HEADER", "Prism monitor mode with 802.11 header"},
	{DLT_PFLOG, "DLT_PFLOG", "OpenBSD pflog"},
	{DLT_FRELAY, "DLT_FRELAY", "Frame Relay"}
};

static Tcl_Condition *dumperMutex;
static unsigned int dumpIndex = 0;

/* I want to check and make sure that I am dealing with the correct type of
 * channel in my procedures.  Tcl_ChannelName lets me do that but that didn't
 * come in until Tcl8.3.2(?)  So I have to use knowledge of Tcl internals to
 * get the same effect for other Tcl versions.  I'm hiding that crap in this
 * macro.
 */
#if (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION < 3) \
    || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION == 3 && \
        TCL_RELEASE_SERIAL ==1)
#	define TCL_CHANNELNAME(x) (x)->typeName
#else
#	define TCL_CHANNELNAME(x) Tcl_ChannelName(x)
#endif

/* Tcl 8.3 introduced Tcl_ReadChars.  For older versions, use Tcl_Read
 * instead.
 */
#if (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION < 3)
#	undef HAS_TCL_READ_CHARS
#else
#	define HAS_TCL_READ_CHARS
#endif

/* Tcl 8.1 introduced the byte array object.  For older versions, use the
 * string object type
 */
#if (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION < 1)
#	undef HAS_TCL_BYTEARRAY
#else
#	define HAS_TCL_BYTEARRAY
#endif

/* FIXME  The man pages indicate that Tcl_GetVersion has been around since
 * 7.5.  The 8.0 that comes with RedHat 6.2 doesn't seem to have this.  For
 * now, if we are < 8.1 I'm going to assume it doesn't exist.
 */
#if (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION < 1)
#	undef HAS_TCL_GETVERSION
#else
#	define HAS_TCL_GETVERSION
#endif

/* If we are on a BSD system, select can't be used on the bpf fd because it
 * won't work properly.  Instead we use an FIONREAD ioctl to the fd to
 * determine if there is any data in the bpf buffers.  This code is contained
 * with the TCLPCAP_NO_SELECT define.
 */
/* In FreeBSD 4.6 and later, this is no longer an issue.  In 4.2, it was still
 * a problem.  I'm leaving this around, in case another platform needs it, but
 * I'll probably remove it in the future.
 */
	/*
#ifdef __FreeBSD__
#	define TCLPCAP_NO_SELECT
#else
#	undef TCLPCAP_NO_SELECT
#endif
	*/

static char const rcsid[] = "$Id: pcap_tcl.c,v 1.41 2005/06/19 01:31:45 todd Exp $";

int Tclpcap_Init (Tcl_Interp *interp)
{
    Tcl_HashTable *hashTable;
#ifdef USE_TCL_STUBS
    if (Tcl_InitStubs(interp, "8.0", 0) == NULL) {
        return TCL_ERROR;
    }
#endif
    if (Tcl_PkgProvide(interp, "Pcap", VERSION) != TCL_OK) {
        return TCL_ERROR;
    }
#define DUMPERKEY "pcapDumpers"
    hashTable = (Tcl_HashTable *) ckalloc((unsigned) sizeof(Tcl_HashTable));
    Tcl_InitHashTable(hashTable, TCL_STRING_KEYS);
    Tcl_SetAssocData(interp, DUMPERKEY, deletePcapDumpers, (ClientData) hashTable);

    Tcl_CreateObjCommand(interp, "pcap::pcap_open", Tcl_pcap_open,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::lookupdev", Tcl_pcap_lookupdev,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::lookupnet", Tcl_pcap_lookupnet,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::getPacket", Tcl_pcap_getPacket,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::packetToString", Tcl_pcap_packetToString,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::printPacket", Tcl_pcap_printPacket,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::savefile", Tcl_pcap_savefile,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::stats", Tcl_pcap_stats,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::datalink", Tcl_pcap_datalink,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::dump_open", Tcl_pcap_dumpOpen,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::dump_close", Tcl_pcap_dumpClose,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::dump", Tcl_pcap_dump,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateObjCommand(interp, "pcap::version", Tcl_pcap_version,
		    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);

    return TCL_OK;
}

void deletePcapDumpers(ClientData clientData, Tcl_Interp *interp)
{
    /* Need to go through the hashtable and close all the dumps that have been
     * opened
     */
    Tcl_HashTable *hashTable;
    Tcl_HashSearch search;
    Tcl_HashEntry *he;
    pcap_dumper_t *dumper;

    hashTable = (Tcl_HashTable *) clientData;
    he = Tcl_FirstHashEntry(hashTable, &search);
    while (he)
    {
	dumper = (pcap_dumper_t *) Tcl_GetHashValue(he);
	pcap_dump_close(dumper);
	he = Tcl_NextHashEntry(&search);
    }
}

int PcapCloseProc (ClientData instanceData, Tcl_Interp *interp)
{
    pcapNode * pInfo = (pcapNode *) instanceData;
    if (pInfo->pcapDumper)
    {
	pcap_dump_close(pInfo->pcapDumper);
    }
    pcap_close(pInfo->pcapPtr);
    Tcl_Free(pInfo->filterString);
    Tcl_Free((ClientData)pInfo);
    return 0;
}

int Tcl_pcap_datalink(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    pcapNode *pnode;
    int i, pdatalink;

    if (pcapGetChannelInstanceData(interp, objc, objv, &pnode) == TCL_ERROR)
	return TCL_ERROR;

    pdatalink = pcap_datalink(pnode->pcapPtr);
    for (i = 0; i < SIZE_DATALINKARRAY; ++i)
    {
	if (pdatalink == datalinkArray[i].dlinkValue)
	{
	    /* Found the datalink entry */
	    Tcl_Obj *matchedDataLink[2];
	    matchedDataLink[0] = Tcl_NewStringObj(
			    datalinkArray[i].shortString, 
			    strlen(datalinkArray[i].shortString));
	    matchedDataLink[1] = Tcl_NewStringObj(
			    datalinkArray[i].longString, 
			    strlen(datalinkArray[i].longString));
	    Tcl_SetObjResult(interp, Tcl_NewListObj(2, matchedDataLink));
	    break;
	}
    }
    if (i == SIZE_DATALINKARRAY)
    {
	/* Don't know about this datalink so return what pcap_datalink gave me */
	Tcl_SetObjResult(interp, Tcl_NewIntObj(pdatalink));
    }
    return TCL_OK;
}

int pcapGetChannelInstanceData(Tcl_Interp *interp, int objc, 
		Tcl_Obj *CONST objv[], pcapNode **pnode)
{
    Tcl_Channel chan;
    char *channelName;
    if (objc != 2)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "pcapChannel");
	return TCL_ERROR;
    }
    channelName = Tcl_GetStringFromObj(objv[1], NULL);
    chan = Tcl_GetChannel(interp, channelName, NULL);
    if (!chan)
    {
	char errmsg[256];
	snprintf(errmsg, 256, "can not find channel named \"%s\"", channelName);
	Tcl_SetErrorCode(interp, "pcap", "no channel", NULL);
	Tcl_AddErrorInfo(interp, errmsg);
	Tcl_SetResult(interp, errmsg, TCL_VOLATILE);
	return TCL_ERROR;
    }

    if (isPcapChannelFromChannel(interp, chan) == TCL_ERROR)
	return TCL_ERROR;

    *pnode = (pcapNode *) Tcl_GetChannelInstanceData(chan);
    return TCL_OK;
}

int Tcl_pcap_stats(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    struct pcap_stat pStat;
    pcapNode *pnode;
    Tcl_Obj *pstatList, *pstatArray[3];

    if (pcapGetChannelInstanceData(interp, objc, objv, &pnode) == TCL_ERROR)
	return TCL_ERROR;
    if (pcap_stats(pnode->pcapPtr, &pStat) != 0)
    {
	Tcl_AddErrorInfo(interp, pcap_geterr(pnode->pcapPtr));
	Tcl_SetErrorCode(interp, "pcap", "noStats", NULL);
	return TCL_ERROR;
    }
    pstatArray[0] = Tcl_NewLongObj(pStat.ps_recv);
    pstatArray[1] = Tcl_NewLongObj(pStat.ps_drop);
    pstatArray[2] = Tcl_NewLongObj(pStat.ps_ifdrop);
    pstatList = Tcl_NewListObj(3, pstatArray);
    Tcl_SetObjResult(interp, pstatList);
    return TCL_OK;
}

int Tcl_pcap_getPacket(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    pcapNode *pnode;
    int result;
    Tcl_Obj *packetList;

    if (pcapGetChannelInstanceData(interp, objc, objv, &pnode) == TCL_ERROR)
	return TCL_ERROR;
    packetList = Tcl_NewObj();
    result = pcapGetPacket(pnode->channel, packetList);
    if (result == TCL_ERROR)
    {
	char errInfoStr[256];
	snprintf(errInfoStr, 256, "Error in getPacket: %s\n", sys_errlist[Tcl_GetErrno()]);
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }
    Tcl_SetObjResult(interp, packetList);
    return TCL_OK;
}

/* This reads the buffer for the channel.  It returns a list.  The first
 * element is a list with some header info from pcap.  The second element is
 * the data read off the wire.
 */
int pcapGetPacket(Tcl_Channel chan, Tcl_Obj *objPtr)
{
    struct pcap_pkthdr packetHeader;
    Tcl_Obj *pcapData[2], *pcapHead[3];
    Tcl_Obj *packetHeaderList;
#ifdef HAS_TCL_READ_CHARS
    Tcl_Obj *readObj = NULL;
#else
    unsigned char *readBuffer = NULL;
#endif
    int readStatus, retval;
    char timeStamp[50];
    unsigned char *packetData;

#ifdef HAS_TCL_READ_CHARS
    readObj = Tcl_NewObj();
    Tcl_IncrRefCount(readObj);
    readStatus = Tcl_ReadChars(chan, readObj, sizeof(struct pcap_pkthdr), 0);
#else
    readBuffer = Tcl_Alloc(sizeof(struct pcap_pkthdr));
    readStatus = Tcl_Read(chan, readBuffer, sizeof(struct pcap_pkthdr)); 
#endif
    if (readStatus == -1)
    {
	/* We don't do anything with the error.  Tcl_pcap_getPacket, which
	 * calls us will let the user know, because they have the interp */
	retval = TCL_ERROR;
	goto exitPcapGetPacket;
    }
    if (readStatus < sizeof(struct pcap_pkthdr))
    {
	/* Either there was nothing to read, or we didn't read enough.  We
	 * just return OK because it really isn't an error.  The app will have
	 * to notice that we returned no data.  This is probably because of an
	 * eof and  [eof chan] should return 1.
	 */
	retval = TCL_OK;
	goto exitPcapGetPacket;
    }
#ifdef HAS_TCL_READ_CHARS
    memcpy(&packetHeader, Tcl_GetByteArrayFromObj(readObj, NULL), 
	    sizeof(struct pcap_pkthdr));
#else
    memcpy(&packetHeader, readBuffer, sizeof(struct pcap_pkthdr));
    Tcl_Free(readBuffer);
#endif
    snprintf(timeStamp, 50, "%lu.%lu", (unsigned long) packetHeader.ts.tv_sec,
	    (unsigned long) packetHeader.ts.tv_usec);
    pcapHead[0] = Tcl_NewStringObj(timeStamp, strlen(timeStamp));
    pcapHead[1] = Tcl_NewIntObj(packetHeader.caplen);
    pcapHead[2] = Tcl_NewIntObj(packetHeader.len);
    packetHeaderList = Tcl_NewListObj(3, pcapHead);
    pcapData[0] = packetHeaderList;
#ifdef HAS_TCL_READ_CHARS
    readStatus = Tcl_ReadChars(chan, readObj, packetHeader.caplen, 0);
#else
    readBuffer = Tcl_Alloc(packetHeader.caplen);
    readStatus = Tcl_Read(chan, readBuffer, packetHeader.caplen); 
#endif
    if (readStatus < packetHeader.caplen)
    {
	retval = TCL_OK;
	goto exitPcapGetPacket;
    }
    if (readStatus == -1)
    {
	/* We don't do anything with the error.  Tcl_pcap_getPacket, which
	 * calls us will let the user know, because they have the interp */
	retval = TCL_ERROR;
	goto exitPcapGetPacket;
    }
#ifdef HAS_TCL_READ_CHARS
#   ifdef HAS_TCL_BYTEARRAY
    packetData = Tcl_GetByteArrayFromObj(readObj, NULL);
#   else
    packetData = Tcl_GetStringFromObj(readObj, NULL);
#   endif
#else
    packetData = readBuffer;
#endif
#ifdef HAS_TCL_BYTEARRAY
    pcapData[1] = Tcl_NewByteArrayObj(packetData, packetHeader.caplen);
#else
    pcapData[1] = Tcl_NewStringObj(packetData, packetHeader.caplen);
#endif

    Tcl_SetListObj(objPtr, 2, pcapData);
    retval = TCL_OK;

exitPcapGetPacket:
#ifdef HAS_TCL_READ_CHARS
    if (readObj) 
	Tcl_DecrRefCount(readObj);
#else
    if (readBuffer) 
	Tcl_Free(readBuffer);
#endif
    return retval;
}

/* This reads a frame from libpcap using pcap_next_ex and stores it in buf.
 * There is no way to tell in advance how much we can read.  We have to deal
 * with a whole frame at a time.  So if we read a frame and it won't fit into
 * buf, the frame gets dropped.  FIXME - We should probably keep track of how
 * many frames get dropped.
 */
int PcapInputProc (ClientData instanceData, char *buf, int bufSize,
		int *errorCodePtr)
{
    u_char *packetData = 0;
    struct pcap_pkthdr *packetHeader = 0;
    pcapNode *pNode = (pcapNode *) instanceData;
    int pcapStatus;
    int selectStatus;

    *errorCodePtr = 0;
#if defined(TCLPCAP_NO_SELECT)
    if (pNode->blocking)
    {
	/* FIXME - I'm not sure I like this.  On FreeBSD, pcap_next would
	 * return immediately instead of blocking.  There is a ioctl to set a
	 * read timeout, but I don't see where it can mean to block forever.
	 * It just seems like there should be a better way.
	 */
	while (!packetData)
	{
	    packetData = (u_char *)pcap_next(pNode->pcapPtr, &packetHeader);
	    if (pNode->offline)
		break;
	}
    } else {
	/* Non-blocking mode */
	if (!pNode->offline) {
	    /* This should work for FreeBSD.  Don't know about others */
	    int numBytes;
	    if (ioctl(pNode->fd, FIONREAD, &numBytes) == -1)
	    {
		/* Handle the error.  Error code is in errno */
	    }
	    if (numBytes > 0)
	    {
		/* There is data to read */
		packetData = (u_char *)pcap_next(pNode->pcapPtr, &packetHeader);
	    }
	} else 
	{
#endif
	    int attemptCount = 1;
	    do {
		pcapStatus = pcap_next_ex(pNode->pcapPtr, &packetHeader, 
			(const u_char **)&packetData);
		if (pcapStatus == 0)
		    continue;
		else
		    break;
	    } while (pNode->blocking || attemptCount++ < 2 );
#if defined(TCLPCAP_NO_SELECT)
	} /* else pNode->offline */
    }  /* else pNode->blocking */
#endif
    if (pcapStatus == -2)
    {
	/* The end of the capture file has been reached so return 0 so EOF is
	 * set 
	 */
	return 0;
    }
    if (pcapStatus == 0)
    {
	/* pcap_next_ex has told us there is no data to read.  If we are in
	 * blocking mode, we should not get here.  So just return 0 indicating
	 * no data to read.  This may cause [eof] to return true
	 *
	 * If we are non-blocking, return -1 with the EAGAIN error to indicate
	 * to the generic layer to try again.
	 */
	if (pNode->blocking)
	{
	    return 0;
	}
	else 
	{
	    *errorCodePtr = EAGAIN;
	    return -1;
	}
    }
    if (pcapStatus == -1)
    {
	/* An error occurred while reading the packet so return 0 for EOF */
	/* Can I give better diagnostics here? */
	return 0;
    }
    
    if (pcapStatus != 1)
    {
	/* This should not happen. pcapStatus should be 1 */
	return 0;
    }

    if(pNode->dumpStatus && pNode->pcapDumper)
    {
	pcap_dump((u_char *)pNode->pcapDumper, packetHeader, packetData);
    }

    if (packetHeader->caplen + sizeof(struct pcap_pkthdr) > bufSize)
    {
	/* No room in the buf for the whole capture */
	return 0;
    }
    memcpy(buf, (void *) packetHeader, sizeof(struct pcap_pkthdr));
    memcpy(buf + sizeof(struct pcap_pkthdr), (const void *)packetData,
		    packetHeader->caplen);

    return packetHeader->caplen + sizeof(struct pcap_pkthdr);
}

/* Return TCL_OK if the given channel is of type pcap.  Return TCL_ERROR if it
 * is not.  This will leave an appropriate error in interp.
 */
int isPcapChannelFromChannel(Tcl_Interp *interp, const Tcl_Channel channel)
{
    if (channel)
    {
	CONST char *chanType;
	chanType = TCL_CHANNELNAME(Tcl_GetChannelType(channel));
	if (strcmp(chanType, "pcap") != 0)
	{
	    Tcl_AddErrorInfo(interp, "wrong channel type");
	    Tcl_SetErrorCode(interp, "pcap", "badChannelType", NULL);
	    Tcl_SetResult(interp, "wrong channel type", TCL_STATIC);
	    return TCL_ERROR;
	}
    } else 
    {
	Tcl_SetErrorCode(interp, "pcap", "badChannelName", NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

int isPcapChannelFromChannelName(Tcl_Interp *interp, const char *channelName)
{
    Tcl_Channel channel;
    channel = Tcl_GetChannel(interp, (char *) channelName, NULL);
    if (!channel)
    {
	char errmsg[256];
	snprintf(errmsg, 256, "can not find channel named \"%s\"", channelName);
	Tcl_SetErrorCode(interp, "pcap", "no channel", NULL);
	Tcl_AddErrorInfo(interp, errmsg);
	Tcl_SetResult(interp, errmsg, TCL_VOLATILE);
	return TCL_ERROR;
    }
    return isPcapChannelFromChannel(interp, channel);
}

/* Implements pcap::savefile pcapChannel [on|off].  If -savefile was specified
 * with fconfigure this will turn on or off dumping of packets to a capture
 * file.  Not specifying on|off will return the current status.
 */
int Tcl_pcap_savefile(ClientData clientData, Tcl_Interp *interp, int objc, 
		Tcl_Obj *CONST objv[])
{
    char *channelName;
    Tcl_Channel channel;
    pcapNode *pnode;
    int dumpFlag;
    if (objc < 2 || objc > 3)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "pcapChannel [on|off]");
	return TCL_ERROR;
    }
    channelName = Tcl_GetStringFromObj(objv[1], NULL);
    if (isPcapChannelFromChannelName(interp, channelName) == TCL_ERROR)
	return TCL_ERROR;

    channel = Tcl_GetChannel(interp, channelName, NULL);
    pnode = (pcapNode *) Tcl_GetChannelInstanceData(channel);
    if (objc == 3)
    {
	if (Tcl_GetBooleanFromObj(interp, objv[2], &dumpFlag) == TCL_ERROR)
	{
	    return TCL_ERROR;
	}
    } else 
    {
	if (pnode->dumpStatus)
	    Tcl_SetResult(interp, "on", TCL_STATIC);
	else
	    Tcl_SetResult(interp, "off", TCL_STATIC);
	return TCL_OK;
    }

    if (dumpFlag)
    {
	pcap_dumper_t *pd;
	if (!pnode->saveFileName)
	{
	    char *errmsg = "fconfigure -savefile has not been called";
	    Tcl_AddErrorInfo(interp, errmsg);
	    Tcl_SetErrorCode(interp, "pcap", "savefile", "name missing", NULL);
	    Tcl_SetResult(interp, errmsg, TCL_VOLATILE);
	    pnode->dumpStatus = 0;
	    pnode->pcapDumper = NULL;
	    return TCL_ERROR;
	}
	pd = pcap_dump_open(pnode->pcapPtr, pnode->saveFileName);
	if (!pd)
	{
	    char *errmsg = "call to pcap_dump_open failed";
	    Tcl_AddErrorInfo (interp, errmsg);
	    Tcl_AddErrorInfo(interp, pcap_geterr(pnode->pcapPtr));
	    Tcl_SetErrorCode(interp, "pcap", "savefile", "pcap_dump_open", NULL);
	    Tcl_SetResult(interp, errmsg, TCL_VOLATILE);
	    pnode->dumpStatus = 0;
	    pnode->pcapDumper = NULL;
	    return TCL_ERROR;
	}
	else
	{
	    pnode->dumpStatus = 1;
	    pnode->pcapDumper = pd;
	}
    } else
    {
	if (pnode->pcapDumper)
	    pcap_dump_close(pnode->pcapDumper);
	pnode->dumpStatus = 0;
	pnode->pcapDumper = NULL;
    }
    return TCL_OK;
}

int PcapOutputProc (ClientData instanceData, CONST char *buf, int toWrite,
		int *errorCodePtr)
{
    /* We will not be supporting this function, but Tcl says we need to
     * provide it.
     */
    *errorCodePtr = EINVAL;
    return -1;
}

#ifdef TCLPCAP_NO_SELECT
typedef struct PcapEvent {
	Tcl_Event header;
	pcapNode *pNode;
} PcapEvent;

int pcapEventHandler (Tcl_Event *evPtr, int flags)
{
    /* No need to check the flags.  I only know about one kind of event */
    Tcl_NotifyChannel(((PcapEvent *)evPtr)->pNode->channel, TCL_READABLE);
    return 1;
}

void PcapEventSetupProc(ClientData clientData, int flags)
{
    Tcl_Time pcapEventTime = {0, 1000};
    int numBytes;
    pcapNode *pNode = (pcapNode *) clientData;
    if (ioctl(pNode->fd, FIONREAD, &numBytes) != -1 && numBytes > 0)
    {
	/* There is still data to read, so don't wait */
	pcapEventTime.usec = 0;
	Tcl_SetMaxBlockTime(&pcapEventTime);
    } else {
	Tcl_SetMaxBlockTime(&pcapEventTime);
    }
}

void PcapEventCheckProc(ClientData clientData, int flags)
{
    int numBytes;
    pcapNode *pNode = (pcapNode *) clientData;
    if (ioctl(pNode->fd, FIONREAD, &numBytes) != -1 && numBytes > 0)
    {
	/* There is data to read.  Post an event. */
	PcapEvent *pEvt = (PcapEvent *)Tcl_Alloc(sizeof(PcapEvent));
	pEvt->pNode = pNode, pEvt->header.proc = pcapEventHandler;
	Tcl_QueueEvent((Tcl_Event *) pEvt, TCL_QUEUE_TAIL);
    }
}
#endif

void PcapWatchProc (ClientData instanceData, int mask)
{
    pcapNode *pNode = (pcapNode *) instanceData;
#ifdef TCLPCAP_NO_SELECT
    if (!pNode->offline) {
	/* Setup an event to fire several times a second and look for data to
	 * read.  We can't use select so we poll.  The event handler will have to
	 * reschedule itself.
	 *
	 * The InputProc should check for data after it reads and if there is
	 * still some, post an event to read some more.  This should allow for
	 * quicker response when a burst of data is received.  If we only read
	 * after a poll, we would limit our speed by the poll rate.
	 *
	 * Make sure the closeProc eliminates the event handler.  Probably need a
	 * flag in pNode.
	 */
	Tcl_CreateEventSource(PcapEventSetupProc, PcapEventCheckProc,
					(ClientData) pNode);
    } else 
#endif
    {
	Tcl_CreateFileHandler(pNode->fd, mask, (Tcl_FileProc *) Tcl_NotifyChannel,
			    (ClientData) pNode->channel);
    }
}

int PcapGetHandleProc (ClientData instanceData, int direction,
		ClientData *handlePtr)
{
    pcapNode *pNode = (pcapNode *) instanceData;
    handlePtr = (ClientData *) pNode->fd;
    return TCL_OK;
}

int PcapBlockModeProc (ClientData instanceData, int mode)
{
    pcapNode *pNode = (pcapNode *) instanceData;
    if (mode == TCL_MODE_BLOCKING)
    {
	pNode->blocking = 1;
    } else if (mode == TCL_MODE_NONBLOCKING)
    {
	pNode->blocking = 0;
    }
    return 0;
}

int PcapSetFilter(Tcl_Interp *interp, pcapNode *pNode, CONST char *value)
{
    char *filterStr;
    struct bpf_program filter;
    if (strcmp(value, "default") == 0)
    {
	filterStr = Tcl_Alloc(strlen("") + 1);
	strcpy(filterStr, "");
    } else
    {
	filterStr = (char *) Tcl_Alloc(strlen(value) + 1);
	strcpy(filterStr, value);
    }
    /* FIXME - We should be calculating the netmask of the local host
     * instead of using 0 in pcap_compile
     */
    if (pcap_compile(pNode->pcapPtr, &filter, filterStr, 1, 0) < 0)
    {
	if (interp)
	{
	    Tcl_AddErrorInfo(interp, pcap_geterr(pNode->pcapPtr));
	    Tcl_SetResult(interp, "Call to pcap_compile failed", TCL_STATIC);
	    Tcl_SetErrorCode(interp, "pcap", "setOption", "-filter", NULL);
	}
	return TCL_ERROR;
    }
    if(pcap_setfilter(pNode->pcapPtr, &filter) < 0)
    {
	pcap_freecode(&filter);
	if (interp)
	{
	    Tcl_AddErrorInfo(interp, pcap_geterr(pNode->pcapPtr));
	    Tcl_SetResult(interp, "Call to pcap_setfilter failed", TCL_STATIC);
	    Tcl_SetErrorCode(interp, "pcap", "setOption", "-filter", NULL);
	}
	return TCL_ERROR;
    }
    pcap_freecode(&filter);
    if (pNode->filterString)
	Tcl_Free(pNode->filterString);
    pNode->filterString = filterStr;
    return TCL_OK;
}

static int PcapSetOptionProc (ClientData instanceData, Tcl_Interp *interp,
		CONST char *optionName, CONST char *value)
{
    pcapNode *pNode = (pcapNode *) instanceData;

    if (strcmp(optionName, "-filter") == 0)
    {
	return PcapSetFilter(interp, pNode, value);
    } else if (strcmp(optionName, "-savefile") == 0)
    {
	/* If we are currently writing to a file, we cannot allow -savefile to
	 * change because it would look like we are writing to one file when
	 * we are really writing to a different one
	 */
	if (pNode->dumpStatus)
	{
	    char *errmsg = "cannot change -savefile while dumping; " \
		    "execute \"pcap::savefile off\" first";
	    Tcl_AddErrorInfo(interp, errmsg);
	    Tcl_SetResult(interp, errmsg, TCL_VOLATILE);
	    Tcl_SetErrorCode(interp, "pcap", "setOption", "-savefile", NULL);
	    return TCL_ERROR;
	}
	if (strcmp(value, "default") == 0)
	{
	    pNode->saveFileName = NULL;
	} else 
	{
	    pNode->saveFileName = Tcl_Alloc(strlen(value) + 1);
	    strcpy(pNode->saveFileName, value);
	}
	return TCL_OK;
    } else
    {
	return Tcl_BadChannelOption(interp, optionName, "filter savefile");
    }
}

static int PcapGetOptionProc (ClientData instanceData, Tcl_Interp *interp,
		CONST char *optionName, Tcl_DString *dsPtr)
{
    pcapNode *pNode = (pcapNode *) instanceData;

    if (optionName)
    {
	if(strcmp(optionName, "-filter") == 0)
	{
	    Tcl_DStringAppendElement(dsPtr, "-filter");
	    Tcl_DStringAppendElement(dsPtr, pNode->filterString);
	} else if(strcmp(optionName, "-savefile") == 0)
	{
	    Tcl_DStringAppendElement(dsPtr, "-savefile");
	    if (pNode->saveFileName)
		Tcl_DStringAppendElement(dsPtr, pNode->saveFileName);
	    else
		Tcl_DStringAppendElement(dsPtr, "");
	} else
	{
	    return Tcl_BadChannelOption(interp, optionName, "filter savefile");
	}
    } else
    {
	Tcl_DStringAppendElement(dsPtr, "-filter");
	Tcl_DStringAppendElement(dsPtr, pNode->filterString);
	Tcl_DStringAppendElement(dsPtr, "-savefile");
	if (pNode->saveFileName)
	    Tcl_DStringAppendElement(dsPtr, pNode->saveFileName);
	else
	    Tcl_DStringAppendElement(dsPtr, "");
    }
    return TCL_OK;
}

/* Use Version 2 channels if available */
#ifdef TCL_CHANNEL_VERSION_2
static Tcl_ChannelType pcapChannelType = {
    "pcap",                     /* Type name. */
    TCL_CHANNEL_VERSION_2,      /* v2 channel */
    PcapCloseProc,              /* Close proc. */
    PcapInputProc,              /* Input proc. */
    PcapOutputProc,             /* Output proc. */
    NULL,                       /* Seek proc. */
    PcapSetOptionProc,          /* Set option proc. */
    PcapGetOptionProc,          /* Get option proc. */
    PcapWatchProc,              /* Initialize notifier. */
    PcapGetHandleProc,          /* Get OS handles out of channel. */
    NULL,                       /* close2proc. */
    PcapBlockModeProc,          /* Set blocking or non-blocking mode.*/
    NULL,                       /* flush proc. */
    NULL                        /* handler proc. */
};
#else
/* Version 1 channel */
static Tcl_ChannelType pcapChannelType = {
    "pcap",                     /* Type name. */
    PcapBlockModeProc,          /* Set blocking or non-blocking mode.*/
    PcapCloseProc,              /* Close proc. */
    PcapInputProc,              /* Input proc. */
    PcapOutputProc,             /* Output proc. */
    NULL,                       /* Seek proc. */
    PcapSetOptionProc,          /* Set option proc. */
    PcapGetOptionProc,          /* Get option proc. */
    PcapWatchProc,              /* Initialize notifier. */
    PcapGetHandleProc,          /* Get OS handles out of channel. */
    NULL                        /* close2proc. */
};
#endif

void createPcapChannel(Tcl_Interp *interp, pcapNode *pcapInfo)
{
    int major = 8, minor = 0;
    Tcl_ChannelType *pcapChannelPtr = &pcapChannelType;
#ifdef HAS_TCL_GETVERSION
    Tcl_GetVersion(&major, &minor, NULL, NULL);
#endif
    pcapInfo->channel = Tcl_CreateChannel(pcapChannelPtr, pcapInfo->pcapName,
		    (ClientData)pcapInfo, TCL_READABLE);
    Tcl_RegisterChannel(interp, pcapInfo->channel);
    if ( (major > 8) || (major == 8 && minor >= 1) )
    {
	/* I think -encoding came in around 8.1 */
	Tcl_SetChannelOption(interp, pcapInfo->channel, "-encoding", "binary");
    }
    Tcl_SetChannelOption(interp, pcapInfo->channel, "-translation", "binary");
}

/* syntax is pcap::pcap_open [options] device 
 *   where options is -offline or -nopromisc
 *      -offline means that device is a capture file
 *      -caplength means the maximum number of bytes to grab
 */
int
Tcl_pcap_open(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    /* by default we open in live mode unless -offline is passed in */
    int offline = 0;
    /* default to promiscuous mode unless -nopromisc is passed */
    int promisc = 1;
    int i;
    char *device, errbuff[PCAP_ERRBUF_SIZE];
    pcap_t *newpcap;
    Tcl_Obj *result;
    char *filterStr = 0;
    int capLength = 1500;
    pcapNode *pNode;

    if (objc <= 1)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "filename|devicename");
	return TCL_ERROR;
    }

    /* Parse the command options */
    for (i = 1; i < objc - 1; ++i)
    {
	char *arg = Tcl_GetStringFromObj(objv[i], NULL);
	if (strcmp(arg, "-offline") == 0)
	{
	    offline = 1;
	} else if (strcmp(arg, "-nopromisc") == 0)
	{
	    promisc = 0;
	} else if (strcmp(arg, "-filter") == 0)
	{
	    filterStr = Tcl_GetStringFromObj(objv[++i], NULL);
	} else if (strcmp(arg, "-caplength") == 0)
	{
	    if (Tcl_GetIntFromObj(interp, objv[++i], &capLength) == TCL_ERROR)
		return TCL_ERROR;
	    if (capLength < 0)
		capLength = 1500;
	}
    }
    device = Tcl_GetStringFromObj(objv[objc-1], NULL);

    if (!filterStr)
    {
	filterStr = (char *)Tcl_Alloc(strlen("default") + 1);
	strcpy(filterStr, "default");
    }

    if (offline)
    {
	newpcap = pcap_open_offline(device, errbuff);
	if (!newpcap)
	{
	    Tcl_SetErrorCode(interp, "pcap", "pcap_open_offline", NULL);
	    Tcl_AddObjErrorInfo(interp, errbuff, strlen(errbuff));
	    Tcl_SetResult(interp, errbuff, TCL_VOLATILE);
	    return TCL_ERROR;
	}
    } else 
    {
	newpcap = pcap_open_live(device, capLength, promisc, 0, errbuff);
	if (!newpcap)
	{
	    Tcl_SetErrorCode(interp, "pcap", "pcap_open_live", NULL);
	    Tcl_AddObjErrorInfo(interp, errbuff, strlen(errbuff));
	    Tcl_SetResult(interp, errbuff, TCL_VOLATILE);
	    return TCL_ERROR;
	}
    }
    pNode = (pcapNode *) Tcl_Alloc(sizeof(pcapNode));
    pNode->offline = offline;
    pNode->fd = pcap_fileno(newpcap);
    pNode->filterString = NULL;
    pNode->saveFileName = NULL;
    pNode->pcapDumper = NULL;
    pNode->dumpStatus = 0;
    if (pNode->fd <= 0)
    {
	pNode->fd = fileno(pcap_file(newpcap));
    }
    snprintf(pNode->pcapName, 20, "pcap%d", pNode->fd);
    pNode->pcapPtr = newpcap;
    createPcapChannel(interp, pNode);

    PcapBlockModeProc ((ClientData) pNode, TCL_MODE_BLOCKING);
    if (PcapSetOptionProc((ClientData) pNode, interp, "-filter", filterStr) == TCL_ERROR)
    {
	/* Hmmm, the user specified filter could not be set.  We could ignore
	 * it and set a default filter, which would get everything, but then
	 * how would the user know?  It seems better to propogate the error
	 * up, which means cleaning up what has been created.
	 */
	Tcl_UnregisterChannel(interp, pNode->channel);
	return TCL_ERROR;
    }

    result = Tcl_NewStringObj(pNode->pcapName, strlen(pNode->pcapName));
    Tcl_SetObjResult(interp, result);

    return TCL_OK;
}


int
Tcl_pcap_lookupdev(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    char *dev, errbuff[PCAP_ERRBUF_SIZE];
    Tcl_Obj *result;
    dev = pcap_lookupdev(errbuff);
    if (!dev)
    {
	Tcl_SetErrorCode(interp, "pcap", "pcap_lookupdev", NULL);
	Tcl_AddObjErrorInfo(interp, errbuff, strlen(errbuff));
	Tcl_SetResult(interp, errbuff, TCL_VOLATILE);
	return TCL_ERROR;
    }
    result = Tcl_NewStringObj(dev, strlen(dev));
    Tcl_SetObjResult(interp, result);
    return TCL_OK;
}

char *ipToStr(bpf_u_int32 ip)
{
    char *str = Tcl_Alloc(sizeof(char) * 16);
    bpf_u_int32 octet1, octet2, octet3, octet4, mask;
    
    ip = htonl(ip);
    mask = 0xff000000;
    octet1 = (ip&mask) >> 24;
    mask = 0x00ff0000;
    octet2 = (ip&mask) >> 16;
    mask = 0x0000ff00;
    octet3 = (ip&mask) >> 8;
    mask = 0x000000ff;
    octet4 = ip&mask;
    sprintf(str, "%d.%d.%d.%d", octet1, octet2, octet3, octet4);
    return str;
}

/* There is a bug in the libpcap shipped with RedHat Linux 6.2.  This causes
 * this to not work as expected because pcap_lookupnet is returning 0 for the
 * net and mask.  When I linked against the latest pcap (0.6.2) it worked
 * fine.
 */
int
Tcl_pcap_lookupnet(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    bpf_u_int32 net = 0, mask = 0;
    char *dev, errbuff[PCAP_ERRBUF_SIZE], *netStr, *maskStr;
    Tcl_Obj *netList, *netInfo[2];
    int retcode;

    if (objc != 2)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "deviceName");
	return TCL_ERROR;
    }
    dev = Tcl_GetStringFromObj(objv[1], NULL);
    retcode = pcap_lookupnet(dev, &net, &mask, errbuff);
    if (retcode == -1)
    {
	Tcl_SetErrorCode(interp, "pcap", "pcap_lookupnet", NULL);
	Tcl_AddObjErrorInfo(interp, errbuff, strlen(errbuff));
	Tcl_SetResult(interp, errbuff, TCL_VOLATILE);
	return TCL_ERROR;
    }
    netStr = ipToStr(net);
    maskStr = ipToStr(mask);
    netInfo[0] = Tcl_NewStringObj(netStr, strlen(netStr));
    netInfo[1] = Tcl_NewStringObj(maskStr, strlen(maskStr));
    /* Tcl_NewStringObj copies these, so we need to delete them now */
    Tcl_Free(netStr); Tcl_Free(maskStr);
    netList = Tcl_NewListObj(2, netInfo);
    Tcl_SetObjResult(interp, netList);
    return TCL_OK;
}


int Tcl_pcap_packetToString(ClientData clientData, Tcl_Interp *interp, int objc, 
		Tcl_Obj *CONST objv[])
{
    Tcl_DString *pktStr;
    Tcl_Obj *pktDataObj;
    unsigned char *pktData;
    int pktLen;
    int i;
#define byteArraySize 256
    unsigned char byteArray[byteArraySize];

    if(objc != 2)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "<packet>");
	return TCL_ERROR;
    }
    pktStr = (Tcl_DString *) Tcl_Alloc(sizeof(Tcl_DString));
    Tcl_DStringInit(pktStr);
    if (Tcl_ListObjIndex(interp, objv[1], 1, &pktDataObj) != TCL_OK)
    {
	return TCL_ERROR;
    }
    if (!pktDataObj)
    {
	return TCL_OK;
    }
#ifdef HAS_TCL_BYTEARRAY
    pktData = Tcl_GetByteArrayFromObj(pktDataObj, &pktLen);
#else
    pktData = Tcl_GetStringFromObj(pktDataObj, &pktLen);
#endif
    for(i = 0; i < pktLen; ++i)
    {
	unsigned int tmp = (pktData[i] << 8) | (pktData[++i]);
	snprintf(byteArray, byteArraySize, "%4.4x ", tmp);
	Tcl_DStringAppend(pktStr, byteArray, strlen(byteArray));
    }

    Tcl_DStringResult(interp, pktStr);
    Tcl_Free( (void *)pktStr);
    return TCL_OK;
}

/* This prints a packet in hex to stdout.  It doesn't print the timestamp
 * or pretty up the output much.  I'm using this for testing now, but we
 * probably need a "real" version
 */
int Tcl_pcap_printPacket(ClientData clientData, Tcl_Interp *interp, int objc, 
		Tcl_Obj *CONST objv[])
{
    Tcl_Obj *pktDataObj;
    unsigned char *pktData;
    int pktLen;
    int i;

    if(objc != 2)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "packet");
	return TCL_ERROR;
    }
    if (Tcl_ListObjIndex(interp, objv[1], 1, &pktDataObj) != TCL_OK)
    {
	return TCL_ERROR;
    }
    if (!pktDataObj)
    {
	return TCL_OK;
    }
#ifdef HAS_TCL_BYTEARRAY
    pktData = Tcl_GetByteArrayFromObj(pktDataObj, &pktLen);
#else
    pktData = Tcl_GetStringFromObj(pktDataObj, &pktLen);
#endif
    for(i = 0; i < pktLen; ++i)
    {
	unsigned int tmp = (pktData[i] << 8) | (pktData[++i]);
	printf("%4.4x ", tmp);
	if ( ((i-1) != 0) && ((i+1)%16 == 0) )
	{
	    printf("\n");
	}
    }
    printf("\n");
    return TCL_OK;
}

int Tcl_pcap_dumpOpen(ClientData clientData, Tcl_Interp *interp, int objc, 
		Tcl_Obj *CONST objv[])
{
    const char *channelName, *dumpFileName;
    pcap_dumper_t *dumper;
    const pcapNode *pnode;
    Tcl_Channel channel;
    Tcl_HashEntry *dumperHashEntry;
    Tcl_HashTable *hashTable;
    char *dumperHashName;
    unsigned int dumperHashNameLen, dumpFileNameLen;
    int new;

    if (objc != 3)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "pcapChannel dumpFileName");
	return TCL_ERROR;
    }

    /* Get the channel information */
    channelName = Tcl_GetString(objv[1]);
    if (isPcapChannelFromChannelName(interp, channelName) == TCL_ERROR)
	return TCL_ERROR;
    channel = Tcl_GetChannel(interp, (char *) channelName, NULL);
    pnode = (pcapNode *) Tcl_GetChannelInstanceData(channel);

    /* Get the dump file information */
    dumpFileName = Tcl_GetStringFromObj(objv[2], &dumpFileNameLen);

    /* Open the dump file.  Report any errors */
    dumper = pcap_dump_open(pnode->pcapPtr, dumpFileName);
    if (dumper == NULL)
    {
	char *errmsg = "call to pcap_dump_open failed";
	Tcl_AddErrorInfo (interp, errmsg);
	Tcl_AddErrorInfo(interp, pcap_geterr(pnode->pcapPtr));
	Tcl_SetErrorCode(interp, "pcap", "dump_open", NULL);
	Tcl_SetResult(interp, errmsg, TCL_VOLATILE);
	return TCL_ERROR;
    }

    /* Store the dumper reference in the hash referenced by the dumpFileName
     * plus an integer index.  The integer index is used in case two files
     * with the same name but different directories are opened at the same
     * time.
     */
    dumperHashNameLen = dumpFileNameLen + 11;
    dumperHashName = ckalloc(dumperHashNameLen);
    if (dumperHashName == NULL)
    {
	Tcl_AddErrorInfo(interp, "Failed to allocate memory for dumperHashName");
	return TCL_ERROR;
    }
    Tcl_MutexLock(dumperMutex);
    dumpIndex++;
    snprintf(dumperHashName, dumperHashNameLen, "%s-%d", dumpFileName, dumpIndex);
    Tcl_MutexUnlock(dumperMutex);
    hashTable = (Tcl_HashTable *) Tcl_GetAssocData(interp, DUMPERKEY, NULL);
    dumperHashEntry = Tcl_CreateHashEntry(hashTable, dumperHashName, &new);
    if (new == 0)
    {
	/* The key already exists!  This should not be possible. */
	pcap_dump_close(dumper);
	Tcl_AddErrorInfo(interp, "Duplicate hash entry exists");
	Tcl_SetErrorCode(interp, "pcap", "dump_open", NULL);
	ckfree(dumperHashName);
	return TCL_ERROR;
    }
    Tcl_SetHashValue(dumperHashEntry, (ClientData) dumper);
    Tcl_SetResult(interp, dumperHashName, TCL_DYNAMIC);

    return TCL_OK;
}

int Tcl_pcap_dumpClose(ClientData clientData, Tcl_Interp *interp, int objc, 
		Tcl_Obj *CONST objv[])
{
    char *dumperName;
    Tcl_HashEntry *dumperEntry;
    pcap_dumper_t *dumper;
    Tcl_HashTable *hashTable;

    if (objc != 2)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "pcap_dumper");
	return TCL_ERROR;
    }

    dumperName = Tcl_GetString(objv[1]);
    hashTable = (Tcl_HashTable *) Tcl_GetAssocData(interp, DUMPERKEY, NULL);
    dumperEntry = Tcl_FindHashEntry(hashTable, dumperName);
    if (dumperEntry == NULL)
    {
	Tcl_AddErrorInfo(interp, "Could not find pcap_dumper");
	return TCL_ERROR;
    }
    dumper = (pcap_dumper_t *) Tcl_GetHashValue(dumperEntry);
    Tcl_DeleteHashEntry(dumperEntry);
    pcap_dump_close(dumper);

    return TCL_OK;
}

int tclHeaderToPcapHeader(Tcl_Interp *interp, Tcl_Obj *headerList, 
	struct pcap_pkthdr *pcapHeader)
{
    Tcl_Obj *tmpObj;
    int listLength;
    unsigned long tv_sec, tv_usec;
    int caplen;
    int len;
    char *timeString;

    if (Tcl_ListObjLength(interp, headerList, &listLength) == TCL_ERROR)
    {
	char *errInfoStr = "The header is not a proper list";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }
    if (listLength != 3)
    {
	char *errInfoStr = "The header is not a proper packet header (too few elements)";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }
    if (Tcl_ListObjIndex(interp, headerList, 1, &tmpObj) == TCL_ERROR)
    {
	char *errInfoStr = "Could not get index 1 from headerList";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }
    if (Tcl_GetIntFromObj(interp, tmpObj, &caplen) == TCL_ERROR)
    {
	char *errInfoStr = "Index 1 is not an integer";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }

    if (Tcl_ListObjIndex(interp, headerList, 2, &tmpObj) == TCL_ERROR)
    {
	char *errInfoStr = "Could not get index 2 from headerList";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }
    if (Tcl_GetIntFromObj(interp, tmpObj, &len) == TCL_ERROR)
    {
	char *errInfoStr = "Index 2 is not an integer";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }

    if (Tcl_ListObjIndex(interp, headerList, 0, &tmpObj) == TCL_ERROR)
    {
	char *errInfoStr = "Could not get index 0 from headerList";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }

    timeString = Tcl_GetString(tmpObj);
    if (sscanf(timeString, "%lu.%lu", &tv_sec, &tv_usec) != 2)
    {
	char *errInfoStr = "Time value has incorrect format";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }

    pcapHeader->caplen = caplen;
    pcapHeader->len = len;
    pcapHeader->ts.tv_sec = tv_sec;
    pcapHeader->ts.tv_usec = tv_usec;
    return TCL_OK;
}

/* syntax: pcap::dump dumper packet
 *    dumper is what is returned from pcap::dump_open
 *    packet is what is returned from pcap::getPacket
 */
int Tcl_pcap_dump(ClientData clientData, Tcl_Interp *interp, int objc, 
		Tcl_Obj *CONST objv[])
{
    char *dumperName;
    Tcl_HashTable *hashTable;
    Tcl_HashEntry *dumperEntry;
    pcap_dumper_t *dumper;
    Tcl_Obj *packetList;
    Tcl_Obj *headerList;
    Tcl_Obj *packet;
    struct pcap_pkthdr pcapHeader;
    u_char *packetData;
    int listLength;

    if (objc != 3)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "pcap_dumper packet");
	return TCL_ERROR;
    }
    dumperName = Tcl_GetStringFromObj(objv[1], NULL);
    hashTable = (Tcl_HashTable *) Tcl_GetAssocData(interp, DUMPERKEY, NULL);
    dumperEntry = Tcl_FindHashEntry(hashTable, dumperName);
    if (dumperEntry == NULL)
    {
	Tcl_AddErrorInfo(interp, "Could not find pcap_dumper");
	return TCL_ERROR;
    }
    dumper = (pcap_dumper_t *) Tcl_GetHashValue(dumperEntry);

    packetList = objv[2];
    if (Tcl_ListObjLength(interp, packetList, &listLength) == TCL_ERROR)
    {
	char *errInfoStr = "The packet is not a proper list";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }
    if (listLength != 2)
    {
	char *errInfoStr = "The packet is not a proper packet (too few elements)";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }
    Tcl_ListObjIndex(interp, packetList, 0, &headerList);
    Tcl_ListObjIndex(interp, packetList, 1, &packet);

    /* Make sure the header has 3 list elements */
    if (Tcl_ListObjLength(interp, headerList, &listLength) == TCL_ERROR)
    {
	char *errInfoStr = "The header is not a proper list";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }
    if (listLength != 3)
    {
	char *errInfoStr = "The header is not a proper packet header (too few elements)";
	Tcl_AddObjErrorInfo(interp, errInfoStr, strlen(errInfoStr));
	return TCL_ERROR;
    }
    if (tclHeaderToPcapHeader(interp, headerList, &pcapHeader) == TCL_ERROR)
    {
	return TCL_ERROR;
    }

#ifdef HAS_TCL_BYTEARRAY
    packetData = Tcl_GetByteArrayFromObj(packet, NULL);
#else
    packetData = Tcl_GetStringFromObj(packet, NULL);
#endif
    pcap_dump((u_char *)dumper, &pcapHeader, packetData);
    return TCL_OK;
}


int Tcl_pcap_version(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    const char * pcap_version = pcap_lib_version();
    Tcl_SetResult(interp, (char *)pcap_version, TCL_VOLATILE);
    return TCL_OK;
}
